package objectmap;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.HashMap;
import java.util.Map.Entry;

import racs.RacsTimer;
import remote.Freeable;
import repository.DataReader;
import repository.ParityFile;
import repository.Repository;

public class ObjectGroup extends Freeable<Integer, String> {
	
	private long groupID;
	private int sizeClass;
	private int parityRepo;
	private HashMap<String, Integer> slots;
	private HashMap<Integer, Boolean> freeSlots;
	private ParityFile pfile;

	@SuppressWarnings("unchecked")
	public ObjectGroup(ObjectMap omap)
	{
		slots = new HashMap<>(32);
		freeSlots = (HashMap<Integer, Boolean>) omap.freeSlots.clone();
	}
	
	@SuppressWarnings("unchecked")
	public ObjectGroup(int sc, long groupID, ObjectMap omap, Integer rk)
	{
		sizeClass = sc;
		this.groupID = groupID;	
		slots = new HashMap<>(32);
		freeSlots = (HashMap<Integer, Boolean>) omap.freeSlots.clone();
		int size = freeSlots.size();
		if(rk != null)
		{
			freeSlots.put(rk, Boolean.FALSE);
			size--;
		}
		int choice = (int) (Math.random()*size);
		//choice = 1;
		int i = -1;
		int r = 0;
		for(Entry<Integer,Boolean> e : freeSlots.entrySet())
		{
			if(e.getValue().booleanValue())
			{
				i++;
			}
			if(i == choice)
			{
				parityRepo = r;
				freeSlots.put(parityRepo, Boolean.FALSE);
				break;
			}
			r++;
		}
		
		size+=100;
	}
	
	public long getID()
	{
		return groupID;
	}
	
	public int getMeta()
	{
		return this.parityRepo;
	}
	
	public boolean removeError(String key)
	{
		return slots.remove(key) != null;
	}
	
	public DataReader degradedGet(ObjectMap omap, int sc, String key) throws Exception
	{
		this.downloadParity(omap, sc);
		
		return pfile.degradedGet(key, omap);
		
	}
	
	public int getFreeRepo(Integer rk, String key) {
		if(rk != null)
		{
			freeSlots.put(rk, Boolean.FALSE);
			slots.put(key, rk);
			return rk.intValue();
		}
		for(Entry<Integer, Boolean> e : freeSlots.entrySet())
		{
			if(e.getValue())
			{
				int ret = e.getKey().intValue();
				e.setValue(Boolean.FALSE);
				slots.put(key, e.getKey());
				return ret;
			}
		}
		return -1;
	}
	
	private void downloadParity(ObjectMap omap, int sc) throws IOException
	{
		if(slots.size() == 0)
		{
			pfile = new ParityFile(sc, this.groupID);
		}
		if(pfile == null)
		{
			pfile = new ParityFile(sc, this.groupID);
			Repository repo = omap.repos.get(this.parityRepo);
			InputStream is;
			try
			{
				is = repo.get(pfile.getName(), "");
			}
			catch(Exception e)
			{
				is = null;
			}
			
			if(is == null)
			{
				pfile = new ParityFile(sc, this.groupID);
			}
			else
			{
				pfile.readFromStream(is);
				is.close();
			}
		}
	}
	
	public void updateParity(Integer rk, String key,
			DataReader data, KeyData kd, ObjectMap omap,
			int sc) throws Exception
	{
		if(slots.size() == 0)
		{
			pfile = new ParityFile(sc, this.groupID);
		}
		if(pfile == null)
		{
			pfile = new ParityFile(sc, this.groupID);
			Repository repo = omap.repos.get(this.parityRepo);
			InputStream is;
			try
			{
				is = repo.get(pfile.getName(), "");
			}
			catch(Exception e)
			{
				is = null;
			}
			
			if(is == null)
			{
				pfile = new ParityFile(sc, this.groupID);
				long before = System.currentTimeMillis();
				pfile.update(key, data, kd);
				long after = System.currentTimeMillis();
				RacsTimer.addData(RacsTimer.PARITY, sc, after-before);
			}
			else
			{
				long before = System.currentTimeMillis();
				pfile.readFromStream(is, key, data, kd);
				long after = System.currentTimeMillis();
				RacsTimer.addData(RacsTimer.PARITY, sc, after-before);
				RacsTimer.addData(RacsTimer.PARITYN, sc, after-before);
				is.close();
			}
		}
		else
		{
			long before = System.currentTimeMillis();
			pfile.update(key, data, kd);
			long after = System.currentTimeMillis();
			RacsTimer.addData(RacsTimer.PARITY, sc, after-before);
			RacsTimer.addData(RacsTimer.PARITYC, sc, after-before);
		}
	}
	
	public void uploadParity(ObjectMap omap) throws Exception
	{
		Repository repo = omap.repos.get(this.parityRepo);
		if(this.slots.size() == 0)
		{
			repo.remove(this.getKey(), "");
		}
		else
		{
			ByteArrayInputStream stream = pfile.getStream();
			long before = System.currentTimeMillis();
			repo.put(this.getKey(), "", stream, stream.available());
			long after = System.currentTimeMillis();
			RacsTimer.addData(RacsTimer.PARITY, this.sizeClass, after-before);
		}
	}
	
	public void uploadParityError(ObjectMap omap) throws Exception
	{
		Repository repo = omap.repos.get(this.parityRepo);
		if(pfile.getCount() == 0)
		{
			repo.remove(this.getKey(), "");
		}
		else
		{
			ByteArrayInputStream stream = pfile.getStream();
			repo.put(this.getKey(), "", stream, stream.available());
		}
	}
	
	public boolean deleteFromParity(String key,
			InputStream data, ObjectMap omap) throws Exception
	{
		int size = slots.size();
		Integer repoID = slots.remove(key);
		if(repoID == null || size == 0)
		{
			return false;
		}
		
		freeSlots.put(repoID, Boolean.TRUE);
		
		Repository repo = omap.repos.get(this.parityRepo);
		pfile = new ParityFile();
		pfile.deleteFromStream(repo.get(getKey(), ""), key, data);
		return true;
	}
	
	public boolean errorDeleteFromParity(String key,
			InputStream data, ObjectMap omap, int repoID) throws Exception
	{
		pfile = new ParityFile();
		Repository repo = omap.repos.get(repoID);
		pfile = new ParityFile();
		InputStream is = repo.get(this.getKey(), "");
		pfile.readFromStream(is);
		is.close();
		
		return pfile.deleteError(data, key);
	}
	
	@Override
	public boolean free(Integer key, String item) {
		if(slots.containsKey(item))
		{
			return false;
		}
		if(key == null)
		{
			for(Entry<Integer,Boolean> entry : freeSlots.entrySet())
			{
				if(entry.getValue())
				{
					return true;
				}
			}
			return false;
		}
		Boolean b = freeSlots.get(key);
		return b == null ? true : b.booleanValue();
	}

	@Override
	public String getKey() {
		return keyFrom(groupID, sizeClass);
	}
	
	public static String keyFrom(long gid, int sc)
	{
		return gid + "+" + sc;
	}

	@Override
	public void parseKey(String key) {
		String[] splt = key.split("\\+");
		groupID = Long.parseLong(splt[0]);
		sizeClass = Integer.parseInt(splt[1]);
	}
	
	
	/* ================================
	 * I/O
	 * ===============================*/
	
	@Override
	protected void writeTo(ObjectOutputStream out) throws IOException {
		out.writeInt(slots.size());
		for(Entry<String, Integer> e : slots.entrySet())
		{
			out.writeUTF(e.getKey());
			out.writeInt(e.getValue());
		}
		out.writeInt(parityRepo);
	}
	
	@Override
	protected void readFrom(ObjectInputStream in) throws IOException {
		int n = in.readInt();
		while(n-- > 0)
		{
			String key = in.readUTF();
			int repo = in.readInt();
			slots.put(key, repo);
			freeSlots.put(repo, Boolean.FALSE);
		}
		parityRepo = in.readInt();
		freeSlots.put(parityRepo, Boolean.FALSE);
	}

	
	//Useless for object groups for now
	@Override
	public String getUniqueKey(String key) {
		return null;
	}

	@Override
	public String parseUniqueKey(String key) {
		// TODO Auto-generated method stub
		return null;
	}
	
}
