package objectmap;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.concurrent.atomic.AtomicLong;

import org.apache.curator.RetryPolicy;
import org.apache.curator.framework.CuratorFramework;
import org.apache.curator.framework.CuratorFrameworkFactory;
import org.apache.curator.framework.recipes.atomic.AtomicValue;
import org.apache.curator.framework.recipes.atomic.DistributedAtomicLong;
import org.apache.curator.framework.recipes.atomic.PromotedToLock;
import org.apache.curator.framework.recipes.locks.InterProcessMutex;
import org.apache.curator.retry.ExponentialBackoffRetry;
import org.apache.zookeeper.KeeperException;
import org.apache.zookeeper.data.Stat;

import racs.RacsTimer;
import remote.LockMap;
import remote.ManagerErrorHandler;
import remote.RemoteMap;
import remote.RemoteReset;
import repository.DataReader;
import repository.FileRepository;
import repository.ParityFile;
import repository.RepoRequest;
import repository.Repository;
import repository.S3Repository;
import tpool.ExecutorThreadPool;

//TODO DELETES
//RACS
//COMPUTE

public class ObjectMap implements ManagerErrorHandler<KeyData>
{
	public final static int size[] = {1, 1<<1, 1<<2, 1<<3, 1<<4, 1<<5, 1<<6, 1<<7, 1<<8, 1<<9,
		1<<10, 1<<11, 1<<12, 1 << 13
	};
	
	private ExecutorThreadPool repoPool;
	private RemoteMap[] groups;
	private RemoteMap keyMap;
	private DistributedAtomicLong[] groupID;
	private LockMap[] groupLocks;
	private LockMap keyLocks;
	private LockMap[] freeListLocks;
	private AtomicLong[] keyID;
	private ObjectFreeList[] freeList;
	private KeyManager managers;
	private int id;
	private CuratorFramework client;
	
	protected final HashMap<Integer, Boolean> freeSlots;
	public HashMap<Integer,Repository> repos;
	
	public ObjectMap(CuratorFramework client, int id) throws Exception
	{
		int cs[] = {64, 32, 32, 32, 32, 32, 16, 16, 8, 8,
			4, 4, 2, 1
		};
		
		this.id = id;
		repos = new HashMap<>();
		repos.put(0,new S3Repository("edu-cornell-cs-racs-gdz4-1-bucket1"));
		repos.put(1,new S3Repository("edu-cornell-cs-racs-gdz4-2-bucket1"));
		repos.put(2,new S3Repository("edu-cornell-cs-racs-gdz4-3-bucket1"));
		
//		repos.put(0,new FileRepository("edu-cornell-cs-racs-gdz4-1-bucket1"));
//		repos.put(1,new FileRepository("edu-cornell-cs-racs-gdz4-2-bucket1"));
//		repos.put(2,new FileRepository("edu-cornell-cs-racs-gdz4-3-bucket1"));
		
		repoPool = new ExecutorThreadPool(16, 16);
		
		groupID = new DistributedAtomicLong[size.length];
		groups = new RemoteMap[size.length];
		groupLocks = new LockMap[size.length];
		freeListLocks = new LockMap[size.length];
		freeList = new ObjectFreeList[size.length];
		keyID = new AtomicLong[size.length];
		
		KeyFactory kf = new KeyFactory();
		
		keyLocks = new LockMap("/KL",client);
		keyMap = new RemoteMap("/KEYS", client);
		managers = new KeyManager("/JOB", keyMap, this,
				kf, this.id, keyLocks);
		
		
		for(int i = 0; i < size.length; i++)
		{
			groups[i] = new RemoteMap("/GR"+i, client);
			groupLocks[i] = new LockMap("/GL"+i,client);
			freeListLocks[i] = new LockMap("/FLL"+i,client);
			freeList[i] = new ObjectFreeList("/FL"+i, client,
					new GroupFactory(i,this), freeListLocks[i],
					groupLocks[i], groups[i], cs[i]);
			groupID[i] = createDL(client, i);
			keyID[i] = new AtomicLong(0);
		}
		
		freeSlots = new HashMap<>();
		for(Integer i : repos.keySet())
		{
			freeSlots.put(i, true);
		}
		
		this.client = client;
	}
	
	private static DistributedAtomicLong createDL(CuratorFramework client, int sc) throws Exception
	{
		RetryPolicy retryPolicy = new ExponentialBackoffRetry(1000, 3);
		DistributedAtomicLong  dl = new DistributedAtomicLong(client, "/DL"+sc, retryPolicy,
				PromotedToLock.builder().lockPath("/DLL"+sc)
				.retryPolicy(retryPolicy).build());
		dl.initialize(0l);
		return dl;
	}
	
	public long getGroupID(int sc) throws Exception
	{
		AtomicValue<Long> result = groupID[sc].increment();
		if(result.succeeded())
		{
			return result.postValue();
		}
		else
		{
			throw new Exception();
		}
	}
	
	public long getKeyID(int sc)
	{
		return keyID[sc].incrementAndGet();
	}
	
	public static int sizeGroup(int nbytes)
	{
		int v = nbytes;
		v--;
		v |= v >> 1;
		v |= v >> 2;
		v |= v >> 4;
		v |= v >> 8;
		v |= v >> 16;
		v++;
		v /= 1024;
		int r; //Bit Twiddling Hacks
		int shift;
		r     = ((v > 0xFFFF) ? 1 : 0) << 4; v >>= r;
		shift = ((v > 0xFF  ) ? 1 : 0) << 3; v >>= shift; r |= shift;
		shift = ((v > 0xF   ) ? 1 : 0) << 2; v >>= shift; r |= shift;
		shift = ((v > 0x3   ) ? 1 : 0) << 1; v >>= shift; r |= shift;
		r |= (v >> 1);
		return r;
	}
	
	public boolean puts(String keys[], DataReader[] datas) throws Exception
	{
		if(keys.length != datas.length)
			return false;
		
		PutsTuple tuples[] = new PutsTuple[keys.length];
		for(int i = 0; i < keys.length; i++)
		{
			tuples[i] = new PutsTuple(datas[i], keys[i]);
		}
		Arrays.sort(tuples);
		int start = 0;
		int end = 1;
		int value = tuples[0].sc;
		for(int i = 1; i < tuples.length; i++)
		{
			if(value == tuples[i].sc)
			{
				end++;
			}
			else
			{
				putsHelper(tuples, start, end);
				start = end++;
				value = tuples[i].sc;
			}
		}
		putsHelper(tuples, start, end);
		return true;
	}
	
	private boolean putsHelper(PutsTuple[] pt, int s, int e) throws Exception
	{
		int sc = pt[s].sc;
		Arrays.sort(pt,s,e,PutsTuple.COMP);
		int count = this.freeSlots.size() - 1;
		InterProcessMutex[] m = new InterProcessMutex[pt.length];
		KeyData[] kd = new KeyData[pt.length];
		KeyData[] okd = new KeyData[pt.length];
		boolean[] first = new boolean[pt.length];
		RepoRequest[] requests = new RepoRequest[pt.length];
		
		
		try
		{
			for(int i = s; i < e;)
			{
				int end = Math.min(e, count+i);
				for(int k = i; k < end; k++)
				{
					m[k] = this.keyLocks.getLock(pt[k].key);
					okd[k] = new KeyData();
					Stat stat = new Stat();
					byte[] bytes = keyMap.get(pt[k].key, stat);
					if(bytes == null)
					{
						okd[k].setVersion(-1);
						first[k] = true;
					}
					else
					{
						okd[k].fromBytes(bytes);
						okd[k].setVersion(stat.getVersion());
						first[k] = false;
					}
				}
				ObjectGroup og = this.freeList[sc].handleEmptyCase(null);
				
				int metaID = og.getMeta();
				
				for(int k = i; k < end; k++)
				{
					int repoID = og.getFreeRepo(null, pt[k].key);
					long keyID = this.getKeyID(sc);
					kd[k] = new KeyData(repoID, og.getID(), id, keyID, pt[i].data.size(), metaID);
					kd[k].setVersion(okd[k].getVersion());
					managers.registerJob(pt[k].key, kd[k]);
					Repository repo = repos.get(repoID);
					RepoRequest request = RepoRequest.putRequest(pt[k].data,
							repo, pt[k].key, "");
					repoPool.execute(request);
					requests[k] = request;
				}
				for(int k = i; k < end; k++)
				{
					og.updateParity(kd[k].repoID,
							pt[k].key, pt[k].data, kd[k], this, sc);
				}
				for(int k = i; k < end; k++)
				{
					if(requests[k].waitOn() == RepoRequest.FAILURE)
					{
						return false;
					}
				}
				
				int version = og.getVersion();
				og.incVersion();
				this.groups[sc].put(og.getKey(), og.getBytes(), version);
				
				og.uploadParity(this);
				this.freeList[sc].putFreeItem(og);
				i = end;
			}
			managers.deregisterMass(pt, kd, okd, first, s, e);
			return true;
		}
		finally
		{
			for(int i = s; i < e; i++)
			{
				if(m[i] != null)
				{
					try
					{
						keyLocks.releaseLock(m[i]);
					}
					catch(Exception ex)
					{
						
					}
				}
			}
		}
	}
	
	public boolean put(String key, DataReader data, Integer rk) throws Exception
	{
		int sc = sizeGroup(data.size());
		ObjectGroup og = null;
		boolean done = false;
		
		InterProcessMutex m =  null;
		InterProcessMutex m2 = null;
		try {
			m = keyLocks.getLock(key);
		} catch (Exception e) { return false; }
		
		try
		{
			boolean first;
			KeyData keyData = new KeyData();
			Stat s = new Stat();
			byte[] bytes = this.keyMap.get(key,s);
			if(bytes == null)
			{
				keyData.setVersion(-1);
				first = true;
			}
			else
			{
				keyData.fromBytes(bytes);
				keyData.setVersion(s.getVersion());
				first = false;
			}

			long before = System.currentTimeMillis();
			og = freeList[sc].getFreeItem(rk, key);
			long after = System.currentTimeMillis();
			RacsTimer.addData(RacsTimer.FREE, sc, after-before);
			m2 = og.getLock();

			int repoID = og.getFreeRepo(rk, key);
			int metaID = og.getMeta();
			if(repoID == -1)
			{
				return false;
			}

			Repository repo = repos.get(repoID);
			if(repo == null)
			{
				return false;
			}

			long keyID = getKeyID(sc);

			KeyData kd = new KeyData(repoID, og.getID(), id, keyID, data.size(), metaID);
			String uploadKey = kd.getUniqueKey(key);
			kd.setVersion(keyData.getVersion());
			managers.registerJob(key, kd);

			RepoRequest request = RepoRequest.putRequest(data, repo, uploadKey, "");
			repoPool.execute(request);
			og.updateParity(repoID, key, data, kd, this, sc);

			before = System.currentTimeMillis();
			int result = request.waitOn();
			after = System.currentTimeMillis();
			RacsTimer.addData(RacsTimer.RECV, sc, after-before);
			if(result == RepoRequest.FAILURE)
			{
				return false;
			}
			
			int version = og.getVersion();
			og.incVersion();
			this.groups[sc].put(og.getKey(), og.getBytes(), version);
			
			og.uploadParity(this);
			
			before = System.currentTimeMillis();
			freeList[sc].putFreeItem(og);
			after = System.currentTimeMillis();
			RacsTimer.addData(RacsTimer.FREE, sc, after-before);
			
			m2 = null;

			managers.deregisterJobs(key, kd, keyData, first);
			done = true;
			return true;
		}
		finally
		{
			try
			{
				if(m != null)
					keyLocks.releaseLock(m);
			}
			catch(Exception e)
			{
				
			}
			try
			{
				if(m2 != null)
					groupLocks[sc].releaseLock(m2);
			}
			catch(Exception e)
			{
				if(done) return true;
				return false;
			}
			
		}
	}
	
	private DataReader degradedGet(String key) throws Exception
	{

		ObjectGroup og = null;
		DataReader file = null;
		InterProcessMutex m =  null;
		InterProcessMutex m2 = null;
		int sc = -1;
		m = keyLocks.getLock(key);
		try
		{
			KeyData keyData = new KeyData();
			Stat s = new Stat();
			byte[] bytes = this.keyMap.get(key,s);
			if(bytes == null)
			{
				throw new Exception();
			}
			else
			{
				keyData.fromBytes(bytes);
				keyData.setVersion(s.getVersion());
			}

			sc = sizeGroup(keyData.size);
			String okey = ObjectGroup.keyFrom(keyData.groupID, sc);
			m2 = this.groupLocks[sc].getLock(okey);
			byte[] groupData = this.groups[sc].get(okey);
			if(groupData == null)
			{
				return null;
			}
			og = new ObjectGroup(this);
			og.fromBytes(groupData);
			og.parseKey(okey);
			file = og.degradedGet(this, sc, key);
			return file;	
		}
		
		finally
		{
			try
			{
				if(m != null)
					keyLocks.releaseLock(m);
			}
			catch(Exception e)
			{
				
			}
			try
			{
				if(m2 != null)
					groupLocks[sc].releaseLock(m2);
			}
			catch(Exception e)
			{
				return file;
			}
			
		}
	}
	
	private DataReader get(String key, boolean lock) throws Exception
	{
		InterProcessMutex m;
		if(lock)
		{
			m = keyLocks.getLock(key);
		}
		else
		{
			m = null;
		}
		try
		{
			KeyData keyData = new KeyData();
			Stat s = new Stat();
			byte[] bytes = this.keyMap.get(key,s);
			if(bytes == null)
			{
				if(m != null)
				{
					throw new Exception("Item not found");
				}
				return null;
			}
			else
			{
				keyData.fromBytes(bytes);
				keyData.setVersion(s.getVersion());
			}
			int repoID = keyData.repoID;
			Repository repo = repos.get(repoID);
			String uploadKey = keyData.getUniqueKey(key);
			try
			{
				return repo.get(uploadKey, "");
			}
			catch(Exception e)
			{
				return null;
			}
		}
		finally
		{
			if(m != null)
				keyLocks.releaseLock(m);
		}
	}
	
	public DataReader get(String key)
	{
		DataReader is = null;
		try
		{
			is = get(key, false);
			if(is == null)
			{
				is = get(key, true);
				if(is == null)
				{
					is = degradedGet(key);
				}
			}
		}
		catch(Exception e)
		{
			
		}
		return is;
	}
	
	public static void main(String[] args) throws Exception
	{
		RetryPolicy retryPolicy = new ExponentialBackoffRetry(1000, 3);
		CuratorFramework client = CuratorFrameworkFactory.newClient("localhost:2181", retryPolicy);
		client.start();
		RemoteReset.reset(client);
		final ObjectMap omap = new ObjectMap(client, 0);
		File f = new File("lyrics1.txt");
		int size = (int)f.length();
		
		FileInputStream fis = new FileInputStream("lyrics1.txt");
		DataReader dr = new DataReader(fis, size);
		boolean dood = false;
		FileInputStream fis2 = new FileInputStream("lyrics3.txt");
		DataReader dr2 = new DataReader(fis2, size);

		omap.put("lyrics.txt", dr, 2);
		omap.put("lyrics2.txt", dr2, 1);
		
		fis = new FileInputStream("lyrics1.txt");
		dr = new DataReader(fis, size);
		dood = false;
		fis2 = new FileInputStream("lyrics3.txt");
		dr2 = new DataReader(fis2, size);
		
		omap.put("lyrics3.txt", dr, 2);
		omap.put("lyrics4.txt", dr2, 1);
		
		InputStream i = omap.get("lyrics2.txt");
		FileOutputStream fos = new FileOutputStream("out.txt");
		byte[] buffer = new byte[4096];
		int size6 = 0;
		while( (size6 = i.read(buffer, 0, buffer.length)) != -1)
		{
			fos.write(buffer, 0, size6);
		}
		fos.flush();
		fos.close();
		
		//omap.remove("lyrics.txt");
		
		System.out.println("Done " + dood);
		System.out.println(omap.freeList[2].cacheHits.get());
		
		
		InputStream is = omap.get("lyrics1.txt");
		System.out.println(is);
	}

	public boolean remove(String key) throws Exception
	{
		boolean done = false;
		InterProcessMutex m1 = keyLocks.getLock(key);
		try
		{
			KeyData keyData = new KeyData();
			Stat s = new Stat();
			byte[] bytes = this.keyMap.get(key,s);
			if(bytes == null)
			{
				return false;
			}
			else
			{
				keyData.fromBytes(bytes);
				keyData.setVersion(s.getVersion());
			}
			String putPath = this.keyMap.putPath(key);
			String jobPath = this.managers.jobPath(key, keyData);
			client.inTransaction().delete().withVersion(keyData.getVersion()).forPath(putPath).
				and().create().forPath(jobPath, keyData.getBytes()).and().commit();
			done = true;
			return true;
		}
		finally
		{
			try
			{
				keyLocks.releaseLock(m1);
			}
			catch(Exception e)
			{
				return done;
			}
		}	
	}
	
	@Override
	public void handleError(String key, KeyData item) throws Exception {
		int sc = sizeGroup(item.size);
		long gid = item.groupID;
		int repoID = item.repoID;
		System.out.println("ERROR FOR " + key);
		Stat stat = keyMap.getStat(key);
		if(!(stat == null) && stat.getVersion() == item.getVersion())
		{
			System.out.println("ERROR VERSION ERROR");
			return;
		}
		Repository repo = this.repos.get(repoID);
		String uploadKey = item.getUniqueKey(key);
		InputStream data = repo.get(uploadKey, "");
		String groupKey = ObjectGroup.keyFrom(gid, sc);
		
		InterProcessMutex m = this.groupLocks[sc].getLock(groupKey);
		try
		{
			Stat s = new Stat();
			ObjectGroup og = new ObjectGroup(this);
			og.fromBytes(this.groups[sc].get(groupKey,s));
			og.setVersion(s.getVersion());
			og.parseKey(groupKey);
			
			try
			{
				if(og.errorDeleteFromParity(key, data, this, item.metaRepo));
				{
					og.uploadParityError(this);
				}
			}
			catch(NullPointerException e)
			{
				e.printStackTrace();
			}
			if(og.removeError(key))
			{
				int version = og.getVersion();
				og.incVersion();
				byte[] bytes = og.getBytes();
				groups[sc].put(groupKey, bytes, version);
			}
			data.close();
			repo.remove(uploadKey, "");
		}
		finally
		{
			this.groupLocks[sc].releaseLock(m);
		}
		
	}
	
	public void printStats()
	{
		int i = 0;
		for(ObjectFreeList list : this.freeList)
		{
			System.out.println(list.cacheHits.get() +" vs. " + list.cacheTries.get() + " for " + i++);
		}
		RacsTimer.printData();
	}
}
