package racs;

import java.io.IOException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.util.Scanner;

import org.apache.curator.RetryPolicy;
import org.apache.curator.framework.CuratorFramework;
import org.apache.curator.framework.CuratorFrameworkFactory;
import org.apache.curator.retry.ExponentialBackoffRetry;

import objectmap.ObjectMap;
import remote.RemoteReset;
import tpool.ExecutorThreadPool;

import com.sun.net.httpserver.HttpExchange;
import com.sun.net.httpserver.HttpHandler;
import com.sun.net.httpserver.HttpServer;

public class RacsHttpServer 
{
	private HttpServer httpServer;
	private ObjectMap omap;
	
	
	public RacsHttpServer(int id, String zk, String addr, int port) throws Exception
	{
		InetSocketAddress inet = new InetSocketAddress(port);
		httpServer = HttpServer.create(inet, 16);
		RetryPolicy retryPolicy = new ExponentialBackoffRetry(1000, 3);
		CuratorFramework client = CuratorFrameworkFactory.newClient(zk, retryPolicy);
		client.start();
		client.blockUntilConnected();
		if(id == 4)
		{
			RemoteReset.reset(client);
			id = 0;
		}
		omap = new ObjectMap(client, id);
		httpServer.setExecutor(new ExecutorThreadPool(32, 32));
		httpServer.createContext("/", new RacsHandler(omap));
		httpServer.start();
	}
	
	public void printData()
	{
		omap.printStats();
	}
	public static void main(String[] args) throws Exception
	{
		int id = 0;
		if(args.length == 1) 
			id = Integer.parseInt(args[0]);
		
		RacsHttpServer s = new RacsHttpServer(id, "localhost:2181", "0.0.0.0", 80);
		System.out.println("SERVER UP");
		Scanner scan = new Scanner(System.in);
		while(scan.hasNext())
		{
			scan.nextLine();
			s.printData();
		}
		
	}
}
