package remote;

import org.apache.curator.framework.CuratorFramework;
import org.apache.curator.utils.EnsurePath;
import org.apache.zookeeper.KeeperException;
import org.apache.zookeeper.data.Stat;

public class RemoteMap
{
	CuratorFramework client;
	
	private String base;
	private static final int size = (1 <<  16) - 1;
	
	public RemoteMap(String basePath, CuratorFramework client) throws Exception
	{ 
		this.client = client;
		try
		{
			client.create().creatingParentsIfNeeded().forPath(basePath);
		}
		catch(KeeperException.NodeExistsException e){}
		base = basePath + "/";
	}
	
	public void put(String key, byte[] data) throws Exception
	{
		int hash = key.hashCode() & size;
		key = base + hash + "/" + key;
		try
		{
			client.create().creatingParentsIfNeeded().forPath(key,data);
		}
		catch(KeeperException.NodeExistsException e)
		{
			client.setData().forPath(key, data);
		}
	}
	
	public String putPath(String key) throws Exception
	{
		int hash = key.hashCode() & size;
		String first = base + hash;
		try {
			client.create().forPath(first);
		} catch (KeeperException.NodeExistsException e) {
		}
		return base + hash + "/" + key;
	}
	
	public void put(String key, byte[] data, int version) throws Exception
	{
		int hash = key.hashCode() & size;
		key = base + hash + "/" + key;
		try
		{	
			client.setData().withVersion(version).forPath(key, data);
			
		}
		catch(KeeperException.NoNodeException e)
		{
			if(version != -1)
			{
				throw new Exception("ERROR VERSION MAP");
			}
			client.create().creatingParentsIfNeeded().forPath(key,data);
		}
	}
	
	public byte[] get(String key) throws Exception
	{
		int hash = key.hashCode() & size;
		key = base + hash + "/" + key;
		try
		{
			return client.getData().forPath(key);
		}
		catch(KeeperException.NoNodeException e)
		{
			
		}
		return null;
	}
	
	public byte[] get(String key, Stat stat) throws Exception
	{
		int hash = key.hashCode() & size;
		key = base + hash + "/" + key;
		try
		{
			return client.getData().storingStatIn(stat).forPath(key);
		}
		catch(KeeperException.NoNodeException e)
		{
			stat.setVersion(-1);
		}
		return null;
	}
	
	public void delete(String key) throws Exception
	{
		int hash = key.hashCode() & size;
		key = base + hash + "/" + key;
		try
		{
			client.delete().forPath(key);
		}
		catch(KeeperException.NoNodeException e)
		{
			
		}
	}
	
	public void delete(String key, int version) throws Exception
	{
		int hash = key.hashCode() & size;
		key = base + hash + "/" + key;
		try
		{
			client.delete().withVersion(version).forPath(key);
		}
		catch(KeeperException.NoNodeException e)
		{
			
		}
	}
	
	public Stat getStat(String key) throws Exception
	{
		int hash = key.hashCode() & size;
		key = base + hash + "/" + key;
		try
		{
			return client.checkExists().forPath(key);
		}
		catch(KeeperException.NoNodeException e)
		{
			return null;
		}
	}
}
