package remote;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;

public abstract class Versionable implements Byteable {

	private int version;
	protected int size = 16;
	
	public abstract String getUniqueKey(String key);
	public abstract String parseUniqueKey(String key);
	
	public final int getVersion()
	{
		return version;
	}
	public final void setVersion(int version)
	{
		this.version = version;
	}
	public final void incVersion()
	{
		version++;
	}
	
	@Override
	public final void fromBytes(byte[] bytes) throws IOException
	{
		ByteArrayInputStream bis = new ByteArrayInputStream(bytes);
		ObjectInputStream in = new ObjectInputStream(bis);
		
		version = in.readInt();
		readFrom(in);
	}
	
	@Override
	public final byte[] getBytes()
	{
		ByteArrayOutputStream bos = new ByteArrayOutputStream(size);
		ObjectOutputStream out;
		try {
			out = new ObjectOutputStream(bos);
			out.writeInt(version);
			writeTo(out);
			out.flush();
			return bos.toByteArray();
		} catch (IOException e) {
			return null;
		}
		
	}
	
	public final static int version(byte[] data) throws IOException
	{
		ByteArrayInputStream bis = new ByteArrayInputStream(data);
		ObjectInputStream in = new ObjectInputStream(bis);
		return in.readInt();
	}
	
	public void writeToStream(ObjectOutputStream oos) throws IOException
	{
		oos.writeInt(version);
		writeTo(oos);
		oos.flush();
	}
	
	public Versionable readFromStream(ObjectInputStream ois) throws IOException
	{
		version = ois.readInt();
		readFrom(ois);
		return this;
	}
	
	protected abstract void writeTo(ObjectOutputStream oos) throws IOException;
	
	protected abstract void readFrom(ObjectInputStream ois) throws IOException;
}
