package repository;

import java.io.IOException;
import java.io.InputStream;

public class DataReader extends InputStream
{
	public byte data[];
	public int readIndex;
	public int writeIndex;
	InputStream is;
	
	public DataReader(InputStream is, int size)
	{
		data = new byte[size];
		this.is = is;
	}

	@Override
	public int read() throws IOException {
		if(readIndex == writeIndex)
		{
			update(readIndex);
			if(readIndex == writeIndex)
			{
				return -1;
			}
		}
		return data[readIndex++];
	}
	
	public int size()
	{
		return data.length;
	}
	
	@Override 
	public int read(byte[] buf, int off, int len) throws IOException
	{
		if(readIndex == writeIndex)
		{
			if(readIndex == data.length) return -1;
			update(readIndex);
		}
		int wi = writeIndex;
		if(wi - readIndex < len)
			len = wi-readIndex;
		
		int ri = readIndex;
		readIndex += len;
		
		System.arraycopy(data, ri, buf, off, len);
		
		return len;
	}
	
	public synchronized int update(int index) throws IOException
	{
		if(index != writeIndex) return writeIndex - index;
		if(writeIndex == data.length) return -1;
		int size = Math.max(is.available(), ParityFile.CHUNK_SIZE);
		if(writeIndex + size > data.length)
		{
			size  = (data.length - writeIndex);
		}
		int bytes = is.read(data, writeIndex, size);
		if(bytes > 0)
		{
			writeIndex += bytes;
		}
		return bytes;
	}
	
	@Override
	public void close()
	{
		try
		{
			is.close();
		}
		catch(Exception e)
		{
			
		}
	}

}
