package repository;

import java.io.InputStream;

public abstract class RepoRequest implements Runnable {
	
	
	public static final int NOT_STARTED = -1;
	
	public static final int IN_PROGRESS = 0;
	
	public static final int SUCCESS = 1;
	
	public static final int FAILURE = 2;
	
	private int status;
	
	protected InputStream data;
	
	protected Repository repo;
	
	protected String key;
	protected String bucket;
	
	RepoRequest(InputStream data, Repository repo, String key, String bucket)
	{
		this.data = data;
		this.repo = repo;
		this.key = key;
		this.bucket = bucket;
		status = NOT_STARTED;
	}
	
	synchronized void done(int status)
	{
		this.status = status;
		notifyAll();
	}
	
	public synchronized int waitOn()
	{
		while(status < SUCCESS)
		{
			try 
			{
				wait();
			} 
			catch (InterruptedException e) {}
		}
		return status;
	}
	
	public final void run()
	{
		status = IN_PROGRESS;
		try
		{
			done(execute());
		}
		catch(Exception e)
		{
			done(FAILURE);
		}
	}
	
	protected abstract int execute() throws Exception;
	
	public static RepoRequest putRequest(DataReader data, Repository repo, String key, String bucket)
	{
		return new PutRequest(data, repo, key, bucket);
	}
	
	public static RepoRequest getRequest(Repository repo, String key, String bucket)
	{
		return new GetRequest(repo, key, bucket);
	}

}
