package tpool;

import java.util.concurrent.Semaphore;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.locks.ReentrantLock;

public class ConcurrentRingBuffer<T> {
	
	private int sizeMask;
	
	private ReentrantLock readLock;
	private ReentrantLock writeLock;
	
	private Semaphore nonFull;
	private Semaphore nonEmpty;
	
	private T[] data;
	
	private int head;
	private int tail;

	
	@SuppressWarnings("unchecked")
	public ConcurrentRingBuffer(int capacity)
	{
		capacity = round(capacity);
		sizeMask = capacity - 1;
		readLock = new ReentrantLock();
		writeLock = new ReentrantLock();
		nonFull = new Semaphore(capacity);
		nonEmpty = new Semaphore(0);
		data = (T[])new Object[capacity];
		head = 0;
		tail = 0;
	}
	
	private static int round(int v)
	{
		v--;
		v |= v >>> 1;
		v |= v >>> 2;
		v |= v >>> 4;
		v |= v >>> 8;
		v |= v >>> 16;
		v++;
		
		return v;
	}
	
	public void put(T t)
	{
		nonFull.acquireUninterruptibly();
		insert(t);
		
	}
	
	public boolean tryPut(T t)
	{
		if(nonFull.tryAcquire())
		{	
			insert(t);
			return true;
		}
		return false;
	}
	
	private void insert(T t)
	{		
		writeLock.lock();
		int tmp = head;
		head = (head + 1) & (sizeMask);
		writeLock.unlock();
		
		data[tmp] = t;
	
		nonEmpty.release();
	}
	
	public T tryTake()
	{
		if(nonEmpty.tryAcquire())
		{
			return remove();
		}
		return null;
	}
	public T take()
	{
		nonEmpty.acquireUninterruptibly();
		return remove();
	}
	
	private T remove()
	{
		T ret;
		
		readLock.lock();
		int tmp = tail;
		tail = (tail + 1) & (sizeMask);
		readLock.unlock();
		
		ret = data[tmp];
		
		nonFull.release();
		
		return ret;
	}
}
