package tpool;

public abstract class ThreadPool<T> 
{
	private int threadCount;
	
	ConcurrentRingBuffer<T> taskQueue;
	
	
	public ThreadPool(int count, int cap)
	{
		threadCount = count;
		taskQueue = new ConcurrentRingBuffer<T>(cap);
		for(int i = 0; i < count; i++)
		{
			getThread().start();
		}
	}
	
	public int getThreadCount()
	{
		return threadCount;
	}
	
	protected PoolThread getThread()
	{
		return new PoolThread();
	}
	
	public final void submit(T arg)
	{
		taskQueue.put(arg);
	}
	
	public final boolean trySubmit(T arg)
	{
		return taskQueue.tryPut(arg);
	}
	
	protected abstract void handle(T t);
	
	protected class PoolThread extends Thread
	{
		@Override
		public final void run()
		{
			while(true)
			{
				try
				{
					handle(taskQueue.take());
				}
				catch(Exception e)
				{
					e.printStackTrace();
				}
			}
		}
	}
}

